import time
from typing import Dict, Any

from atlas_engine_client.core.api import UserTaskQuery

from robot.api import logger

from ._fields_helper import filter_kwargs_for_dataclass
from ._retry_helper import retry_on_exception


class UserTaskKeyword:

    def __init__(self, client, **kwargs):
        self._client = client

        self._max_retries = kwargs.get('max_retries', 5)
        self._backoff_factor = kwargs.get('backoff_factor', 2)
        self._delay = kwargs.get('delay', 0.1)

    @retry_on_exception
    def get_user_task_by(self, **kwargs):

        logger.debug(kwargs)

        current_retry = 0
        current_delay = float(kwargs.get('delay', self._delay))
        backoff_factor = float(kwargs.get('backoff_factor', self._backoff_factor))
        max_retries = int(kwargs.get('max_retries', self._max_retries))

        local_kwargs = filter_kwargs_for_dataclass(UserTaskQuery, kwargs)

        query = UserTaskQuery(**local_kwargs)

        logger.info(query)

        while True:
            user_tasks = self._client.user_task_get(query)

            logger.info(user_tasks)

            if len(user_tasks) >= 1:
                user_task = user_tasks[0]
            else:
                user_task = {}

            if user_task:
                break
            else:
                time.sleep(current_delay)
                current_retry = current_retry + 1
                current_delay = current_delay * backoff_factor
                if current_retry > max_retries:
                    break
                logger.info(f" Retry count: {current_retry} of {max_retries}; delay: {current_delay} and backoff_factor: {backoff_factor}")

        return user_task

    @retry_on_exception
    def finish_user_task(self, user_task_instance_id: str, payload: Dict[str, Any], **kwargs):
        self._client.user_task_finish(user_task_instance_id, payload)



import time
from urllib3.exceptions import NewConnectionError

from atlas_engine_client.core.api import Client

from .keywords import DeployKeyword
from .keywords import EmptyTaskKeyword
from .keywords import EngineKeyword
from .keywords import ExternalTaskKeyword
from .keywords import ManualTaskKeyword
from .keywords import ProcessInstanceKeyword
from .keywords import SendKeyword
from .keywords import StartKeyword
from .keywords import UserTaskKeyword

from robot.api import logger

from .docker_handler import DockerHandler


class ProcessCubeLibrary(DeployKeyword, EmptyTaskKeyword, 
    EngineKeyword, ExternalTaskKeyword, ManualTaskKeyword, 
    ProcessInstanceKeyword, SendKeyword, StartKeyword, 
    UserTaskKeyword):

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'

    def __init__(self, **kwargs):
        self._engine_url = kwargs.get('engine_url', None)
        self._self_hosted_engine = kwargs.get('self_hosted_engine', None)
        self._docker_options = kwargs.get('docker_options', {})

        self._max_retries = int(kwargs.get('max_retries', 5))
        self._backoff_factor = int(kwargs.get('backoff_factor', 2))
        self._delay = float(kwargs.get('delay', 0.1))
        self._worker_id = kwargs.get('worker_id', "robot_framework")
        readyness_retries = int(kwargs.get('readyness_retry', 20))

        self._client = self._create_client(**kwargs)
        self._kwargs = kwargs


        super(ProcessCubeLibrary, self).__init__(self._client, **kwargs)

        # check if engine is ready
        self._check_engine_readyness(readyness_retries)


    def log_start_parameters(self):
        kwargs_repr = [f"{k}={v!r}" for k, v in self._kwargs.items()]
        logger.info(f"start library with kwargs {kwargs_repr}")

    def _create_client(self, **kwargs):
        engine_url = kwargs.get('engine_url', None)
        self_hosted_engine = kwargs.get('self_hosted_engine', None)

        client = None

        if engine_url:
            logger.console(f"Connecting to engine '{engine_url}'.")
            client = Client(engine_url)
        elif self_hosted_engine == 'docker':
            docker_options = kwargs.get('docker_options', {})

            self._docker_handler = DockerHandler(**docker_options)
            self._docker_handler.start()

            engine_url = self._docker_handler.get_engine_url()
            #engine_url = "http://localhost:1234"

            logger.console(f"Connecting to engine '{engine_url}'.")
            client = Client(engine_url)
        else:
            raise TypeError(
                "No 'engine_url' or 'self_hosted_engine' parameter provided.")

        return client

    def _check_engine_readyness(self, readyness_retries):
        current_retry = 0
        while True:
            try:
                value = self._client.info()
            
                return value
                
            except Exception as e:
                time.sleep(1)
                current_retry = current_retry + 1
                
                if current_retry > readyness_retries:
                    raise Exception(f"Calling check-engine-readyness with readyness_retries {readyness_retries} without success.") from e

                logger.info(f"[check-engine-readyness] Retry count: {current_retry}; delay.")

    def engine_shutdown(self):
        if self._docker_handler:
            self._docker_handler.shutown()




import time
from typing import Dict, Any

from atlas_engine_client.core.api import FetchAndLockRequestPayload
from atlas_engine_client.core.api import FinishExternalTaskRequestPayload

from robot.api import logger

from ._retry_helper import retry_on_exception


class ExternalTaskKeyword:

    def __init__(self, client, **kwargs):
        self._client = client

    @retry_on_exception
    def get_external_task(self, topic: str, options: dict = {}, **kwargs):

        request = FetchAndLockRequestPayload(
            worker_id=self._worker_id,
            topic_name=topic,
            max_tasks=1
        )

        logger.info(f"get task with {request}")

        current_retry = 0
        current_delay = float(kwargs.get('delay', self._delay))
        backoff_factor = float(kwargs.get('backoff_factor', self._backoff_factor))
        max_retries = int(kwargs.get('max_retries', self._max_retries))

        while True:
            external_tasks = self._client.external_task_fetch_and_lock(request)

            logger.info(external_tasks)

            if len(external_tasks) == 1:
                external_task = external_tasks[0]
            else:
                external_task = {}

            if external_task:
                break
            else:
                time.sleep(current_delay)
                current_retry = current_retry + 1
                current_delay = current_delay * backoff_factor
                if current_retry > max_retries:
                    break
                logger.info(f" Retry count: {current_retry} of {max_retries}; delay: {current_delay} and backoff_factor: {backoff_factor}")

        return external_task

    @retry_on_exception
    def finish_external_task(self, external_task_id: str, result: Dict[str, Any], **kwargs):
        request = FinishExternalTaskRequestPayload(
            worker_id=self._worker_id,
            result=result
        )

        logger.info(f"finish task with {request}")

        self._client.external_task_finish(external_task_id, request)



from dataclasses import fields
from threading import local

def filter_kwargs_for_dataclass(dataclassType, kwargs: dict) -> dict:
    local_kwargs = kwargs.copy()
    
    my_fields = fields(dataclassType)
    field_names = [field.name for field in my_fields]
    
    for key in kwargs:
        if key not in field_names:
            del local_kwargs[key] 


    return local_kwargs



from ._retry_helper import retry_on_exception


class EngineKeyword:

    def __init__(self, client, **kwargs):
        self._client = client

    @retry_on_exception
    def get_engine_info(self, **kwargs):
        info = self._client.info()

        return info
